@@ -3,31 +3,81 @@ package Coro::ProcessPool::Util;
 use strict;
 use warnings;
 use Carp;
+use Config;
 use Const::Fast;
-use Storable;
-use Coro::Storable qw(nfreeze thaw);
-use MIME::Base64   qw(encode_base64 decode_base64);
+use Storable        qw(freeze thaw);
+use MIME::Base64    qw(encode_base64 decode_base64);
+use String::Escape  qw(backslash);
+use Sereal::Encoder qw(sereal_encode_with_object SRL_SNAPPY);
+use Sereal::Decoder qw(sereal_decode_with_object);
 
 use base qw(Exporter);
-our @EXPORT_OK = qw(cpu_count encode decode $EOL);
+our @EXPORT_OK = qw(
+    get_command_path
+    get_args
+    cpu_count
+    encode
+    decode
+    $EOL
+);
 
 const our $EOL => "\n";
+const our $DEFAULT_READ_TIMEOUT => 0.1;
+
+my $ENCODER = Sereal::Encoder->new();
+my $DECODER = Sereal::Decoder->new();
 
 sub encode {
-    no warnings 'once';
-    local $Storable::Deparse    = 1;
-    local $Storable::forgive_me = 1;
-    my $ref  = shift or croak 'encode: expected reference';
-    my $data = nfreeze($ref);
-    return encode_base64($data, '');
+    my ($id, $info, $data) = @_;
+    $data ||= [];
+
+    my $package = {
+        id   => $id,
+        data => $data,
+        info => undef,
+        code => undef,
+    };
+
+    if (ref $info && ref $info eq 'CODE') {
+        no warnings 'once';
+        local $Storable::Deparse    = 1;
+        local $Storable::forgive_me = 1;
+        $package->{code} = freeze($info);
+    } else {
+        $package->{info} = $info;
+    }
+
+    my $pickled = sereal_encode_with_object($ENCODER, $package);
+    return encode_base64($pickled, '');
 }
 
 sub decode {
-    no warnings 'once';
-    local $Storable::Eval = 1;
     my $line = shift or croak 'decode: expected line';
-    my $data = decode_base64($line) or croak 'invalid data';
-    return thaw($data);
+    my $pickled = decode_base64($line);
+    my $package = sereal_decode_with_object($DECODER, $pickled);
+
+    my ($id, $info, $data) = @{$package}{qw(id info data)};
+
+    if ($package->{code}) {
+        no warnings 'once';
+        local $Storable::Eval = 1;
+        $info = thaw($package->{code});
+    }
+
+    return ($id, $info, $data);
+}
+
+sub get_command_path {
+    my $perl = $Config{perlpath};
+    my $ext  = $Config{_exe};
+    $perl .= $ext if $^O ne 'VMS' && $perl !~ /$ext$/i;
+    return $perl;
+}
+
+sub get_args {
+    my @inc  = map { sprintf('-I%s', backslash($_)) } @INC;
+    my $cmd  = q|-MCoro::ProcessPool::Worker -e 'Coro::ProcessPool::Worker->new->run'|;
+    return join ' ', @inc, $cmd;
 }
 
 #-------------------------------------------------------------------------------
